#define WIN32_LEAN_AND_MEAN
#define WIN32_EXTRA_LEAN
#include <windows.h>
#include <GL/gl.h>
#include "glext.h"

#ifdef _DEBUG
#include <stdio.h>
#endif

#define TINY
//#define FULL


#ifdef _DEBUG

// #define WIDTH       1280
// #define HEIGHT      800
// #define HALF_HEIGHT 400
// #define RATIO       1.6

#define WIDTH       1920
#define HEIGHT      1440
#define HALF_HEIGHT 720
#define RATIO       4./3

#else

// #define WIDTH       1280
// #define HEIGHT      800
// #define HALF_HEIGHT 400
// #define RATIO       1.6

// #define WIDTH       1680
// #define HEIGHT      1050
// #define HALF_HEIGHT 525
// #define RATIO       1.6

// #define WIDTH       1920
// #define HEIGHT      1200
// #define HALF_HEIGHT 600
// #define RATIO       1.6

#define WIDTH       2560
#define HEIGHT      1600
#define HALF_HEIGHT 800
#define RATIO       1.6

// #define WIDTH       800
// #define HEIGHT      600
// #define HALF_HEIGHT 300
// #define RATIO       4./3

// #define WIDTH       1024
// #define HEIGHT      768
// #define HALF_HEIGHT 384
// #define RATIO       4./3

// #define WIDTH       1600
// #define HEIGHT      1200
// #define HALF_HEIGHT 600
// #define RATIO       4./3

// #define WIDTH       2048
// #define HEIGHT      1536
// #define HALF_HEIGHT 768
// #define RATIO       4./3

// #define WIDTH       1280
// #define HEIGHT      720
// #define HALF_HEIGHT 360
// #define RATIO       16./9

// #define WIDTH       1920
// #define HEIGHT      1080
// #define HALF_HEIGHT 540
// #define RATIO       16./9

// #define WIDTH       1280
// #define HEIGHT      1024
// #define HALF_HEIGHT 512
// #define RATIO       1.25

#endif

#include "shader.h"

const static PIXELFORMATDESCRIPTOR pfd =
{
	0, 0, PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

#ifdef FULL
static DEVMODE g_ScreenSettings =
{
	{ 0 }, 0, 0, sizeof(DEVMODE), 0, DM_PELSWIDTH | DM_PELSHEIGHT | DM_BITSPERPEL, { 0 }, 0, 0, 0, 0, 0,
	{ 0 }, 0, 32, WIDTH, HEIGHT, { 0 }, 0, 0, 0, 0, 0, 0, 0, 0, 0
};
#endif


#ifdef TINY

static const float g_Coords[] =
{
	+0.30078125f, +0.02343750f,
	-0.82421875f, +0.18359375f,
	+0.07031250f, -0.62109375f,
	-0.07421875f, -0.66015625f,
	-1.65625000f, +0.00000000f,
};

#else

float g_Coords[] = { 0.0f, -2.0f, 0.0f, 0.0f };

__forceinline bool MandelCount(const float x, const float y)
{
	float cx = x;
	float cy = y;

	float sqx, sqy;

	unsigned char n = 0;
	do
	{
		sqy = cy * cy; // y^2

		cy += cy;
		cy *= cx;
		cy += y; // 2 * x * y + y0
		sqx = cx * cx; // x^2
		cx = sqx - sqy + x;   // x^2 - y^2 + x0
		--n;
		if (n == 0)
			return false;

	} while (sqx + sqy < 4.0f);

	return true;
}
#endif


#ifdef _DEBUG

bool g_Keys[256] = { false };

LRESULT CALLBACK WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
	case WM_KEYDOWN:
		if (wParam == VK_ESCAPE)
		{
			PostQuitMessage(0);
			return 0;
		}
		g_Keys[wParam & 0xFF] = true;
		break;
	case WM_KEYUP:
		g_Keys[wParam & 0xFF] = false;
		break;
	case WM_CLOSE:
	case WM_DESTROY:
		PostQuitMessage(0);
		return 0;
	};

	return DefWindowProc(hwnd, msg, wParam, lParam);
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
#else
// Take away prolog and epilog, then put a minial prolog back manually with assembly below. The function never returns so no epilog is necessary.
__declspec(naked) void winmain()
#endif
{
#ifdef _DEBUG
	WNDCLASS wc;
	memset(&wc, 0, sizeof(wc));
	wc.style         = CS_OWNDC | CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc   = WndProc;
	wc.hInstance     = hPrevInstance;
	wc.lpszClassName = "GL";

	if (!RegisterClass(&wc))
		return 0;

	DWORD style = WS_VISIBLE | WS_CAPTION | WS_SYSMENU | WS_CLIPSIBLINGS | WS_CLIPCHILDREN;

	RECT rec = { 0, 0, WIDTH, HEIGHT };
	AdjustWindowRect( &rec, style, 0 );

	int w = rec.right - rec.left;
	int h = rec.bottom - rec.top;
	int x = (GetSystemMetrics(SM_CXSCREEN) - w) / 2;
	int y = (GetSystemMetrics(SM_CYSCREEN) - h) / 2;

	// Create window and initialize OpenGL
	HWND hwnd = CreateWindowEx(WS_EX_APPWINDOW | WS_EX_WINDOWEDGE, "GL", "Humus", style, x, y, w, h, 0, 0, hPrevInstance, 0);
#else
	// Prolog
	__asm enter 0x10, 0;
	__asm pushad;


	{ // Extra scope to make compiler accept the __decalspec(naked) with local variables

#ifdef FULL
		// Set fullscreen mode
		ChangeDisplaySettings(&g_ScreenSettings, CDS_FULLSCREEN);
#endif

		// Create window and initialize OpenGL
		HWND hwnd = CreateWindow("edit", 0, WS_POPUP | WS_VISIBLE | WS_MAXIMIZE, 0, 0, 0, 0, 0, 0, 0, 0);
#endif

		HDC hdc = GetDC(hwnd);

		SetPixelFormat(hdc, ChoosePixelFormat(hdc, &pfd), &pfd);
		wglMakeCurrent(hdc, wglCreateContext(hdc));

		ShowCursor(FALSE);

		// Create shader
		GLuint prog = ((PFNGLCREATEPROGRAMPROC) wglGetProcAddress("glCreateProgram"))();
		GLuint shader = ((PFNGLCREATESHADERPROC) wglGetProcAddress("glCreateShader"))(GL_FRAGMENT_SHADER);
		((PFNGLSHADERSOURCEPROC) wglGetProcAddress("glShaderSource"))(shader, 1, &g_ShaderCode, 0);
		((PFNGLCOMPILESHADERPROC) wglGetProcAddress("glCompileShader"))(shader);
		((PFNGLATTACHSHADERPROC) wglGetProcAddress("glAttachShader"))(prog, shader);
		((PFNGLLINKPROGRAMPROC) wglGetProcAddress("glLinkProgram"))(prog);
		((PFNGLUSEPROGRAMPROC) wglGetProcAddress("glUseProgram"))(prog);

#ifdef _DEBUG

		PFNGLGETSHADERIVPROC glGetShaderiv = (PFNGLGETSHADERIVPROC) wglGetProcAddress("glGetShaderiv");

		//PFNGLGETPROGRAMINFOLOGPROC glGetProgramInfoLog = (PFNGLGETPROGRAMINFOLOGPROC) wglGetProcAddress("glGetProgramInfoLog");
		PFNGLGETSHADERINFOLOGPROC glGetShaderInfoLog = (PFNGLGETSHADERINFOLOGPROC) wglGetProcAddress("glGetShaderInfoLog");

		char str[512];
		//glGetProgramInfoLog(prog, sizeof(str), NULL, str);
		glGetShaderInfoLog(shader, sizeof(str), NULL, str);
		OutputDebugString("\n\n");
		OutputDebugString(str);
		OutputDebugString("\n\n\n");

		GLint compiled;
		glGetShaderiv(shader, GL_COMPILE_STATUS, &compiled);
		if (!compiled)
		{
			MessageBox(hwnd, str, "Error", MB_OK);
			return -1;
		}

		sprintf_s(str, "Shader:\n\n%s\n\nLength: %d\n\n", g_ShaderCode, strlen(g_ShaderCode));
		OutputDebugString(str);

#endif

		int last_t, tt = 0xFFFFFFFF;
#ifdef TINY
		int last_i, i = 0;
#endif

		for (;;)
		{
			int rnd = GetTickCount();

			last_t = tt;
			tt = (rnd & 8191);

			if ((unsigned int) tt < (unsigned int) last_t)
			{
#ifdef TINY
				last_i = i;
				++i;
				if (i >= (sizeof(g_Coords) / sizeof(g_Coords[0]) / 2))
					i = 0;
#else
				float cos_a, sin_a;

				__asm
				{
					fild rnd
					fsincos
					fstp cos_a
					fstp sin_a
				}

				float r = 4.0f;

				float x, y;

				do
				{
					r -= 1.0f / 8192.0f;
					x = r * cos_a - 1.0f/*0.75f*/;
					y = r * sin_a;
				} while (MandelCount(x, y));

				g_Coords[0] = g_Coords[1];
				g_Coords[2] = g_Coords[3];
				g_Coords[1] = x;
				g_Coords[3] = y;
#endif
			}

			float t1 = tt * (1.0f / 8192.0f);

			float f = 4 * (t1 - t1 * t1);
			f *= f;
			f *= f;
			f *= f;
			f *= f;

			float s = t1;
			s = s * s * (3 - s - s);
			s = s * s * (3 - s - s);
			s = s * s * (3 - s - s);
			s = s * s * (3 - s - s);

			glTexCoord3f(
#ifdef TINY
				g_Coords[2*last_i+0] - s * g_Coords[2*last_i+0] + s * g_Coords[2*i+0],
				g_Coords[2*last_i+1] - s * g_Coords[2*last_i+1] + s * g_Coords[2*i+1],
#else
				g_Coords[0] - s * g_Coords[0] + s * g_Coords[1],
				g_Coords[2] - s * g_Coords[2] + s * g_Coords[3],
#endif
				f + (1.0f / 8192.0f));

			glRects(-1, -1, 1, 1);
			SwapBuffers(hdc);


#ifdef _DEBUG
			bool done = false;

			MSG msg;
			while (PeekMessage(&msg, 0, 0, 0, PM_REMOVE))
			{
				if (msg.message == WM_QUIT)
					done = true;

				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}

			if (done)
				break;
#else
			if (GetAsyncKeyState(VK_ESCAPE))
				break;
#endif

		}
#ifndef _DEBUG
	}
#endif

	ExitProcess(0);
}
